from dataclasses import dataclass
from pathlib import Path
from typing import Iterator, List
from datetime import datetime

import numpy as np
from numpy._typing import NDArray

from .sm4_object_types import StringData

from .sm4_file import (
    Sm4FileAll,
    RhkPageType,
    RhkLineType,
    RhkImageType,
    RhkScanType,
    Sm4PageHeaderDefault,
)


@dataclass()
class Sm4Channel:
    """Class with information about a channel

    page_type: Type of page/channel
    line_type: Type of line
    datetime: Datetime of measurement
    xres: Resolution in x, e.g. number of pixels for images
    yres: Resolution in y, e.g. number of pixels for images
    image_type: Type of image
    scan_type: Type of scan
    scan_direction: Direction of scanning
    xsize: Physical size of e.g. image in x (in m)
    ysize: Physical size of e.g. image in y (in m)
    z_scale:
    x_offset:
    y_offset:
    z_offset:
    period: Acquisition time of a single data point (in s)
    bias: Bias voltage (in V)
    current: Tunneling current (in A)
    angle: Scan angle (in deg)
    data: Mesuremet data
    """

    label: str
    page_type: RhkPageType
    line_type: RhkLineType
    datetime: datetime
    xres: int
    yres: int
    image_type: RhkImageType
    scan_type: RhkScanType
    scan_direction: str
    xsize: float
    ysize: float
    z_scale: float
    x_offset: float
    y_offset: float
    z_offset: float
    period: float
    bias: float
    current: float
    angle: float
    data: NDArray[np.float32]


class Sm4:
    def __init__(self, filepath: str):
        self.filepath = filepath
        sm4file = Sm4FileAll(filepath)
        self.prm_str = sm4file.file_header.prm.prm_data
        self._channels: List[Sm4Channel] = []
        for ch in sm4file.pages:
            if isinstance(ch.header, Sm4PageHeaderDefault):
                ch_datetime = None
                for i in ch.header.page_header_objects:
                    if type(i) == StringData:
                        month_str, day_str, year_str = i.date.split("/")
                        year_str = "20" + year_str
                        month, day, year = (
                            int(month_str),
                            int(day_str),
                            int(year_str),
                        )
                        hour, min, sec = [int(x) for x in i.time.split(":")]
                        ch_datetime = datetime(
                            year, month, day, hour, min, sec
                        )

                # fallback to file stats
                if ch_datetime is None:
                    file_datetime = Path(filepath).stat().st_ctime
                    ch_datetime = datetime.fromtimestamp(file_datetime)

                self._channels.append(
                    Sm4Channel(
                        label=ch.label,
                        page_type=ch.header.page_type,
                        line_type=ch.header.line_type,
                        datetime=ch_datetime,
                        xres=ch.header.x_size,
                        yres=ch.header.y_size,
                        image_type=ch.header.image_type,
                        scan_type=ch.header.scan_type,
                        scan_direction=ch.header.scan_type.direction(),
                        xsize=abs(ch.header.x_scale * ch.header.x_size),
                        ysize=abs(ch.header.y_scale * ch.header.y_size),
                        z_scale=ch.header.z_scale,
                        x_offset=ch.header.x_offset,
                        y_offset=ch.header.y_offset,
                        z_offset=ch.header.z_offset,
                        period=ch.header.period,
                        bias=ch.header.bias,
                        current=ch.header.current,
                        angle=ch.header.angle,
                        data=ch.data.data,
                    )
                )

    def __repr__(self) -> str:
        return repr(self._channels)

    def __len__(self) -> int:
        return len(self._channels)

    def __getitem__(self, idx: int) -> Sm4Channel:
        return self._channels[idx]

    def __setitem__(self, idx: int, item: Sm4Channel) -> None:
        self._channels[idx] = item

    def __delitem__(self, idx: int) -> None:
        del self._channels[idx]

    def __iter__(self) -> Iterator[Sm4Channel]:
        return iter(self._channels)

    def save_prm(self, out_file: str) -> None:
        with open(out_file, "w") as f:
            f.write(self.prm_str)

    def topography_channels(self) -> List[Sm4Channel]:
        return [
            ch
            for ch in self
            if ch.page_type == RhkPageType.RHK_PAGE_TOPOGRAPHIC
        ]

    def current_channels(self) -> List[Sm4Channel]:
        return [
            ch for ch in self if ch.page_type == RhkPageType.RHK_PAGE_CURRENT
        ]



from __future__ import annotations
from typing import List, Union, Optional
from io import BufferedReader
from dataclasses import dataclass, field
from enum import Enum

import numpy as np

from .cursor import Cursor
from .sm4_object_types import (
    ApiInfo,
    ImageDriftHeader,
    ImageDriftData,
    PageData,
    Prm,
    SpecDriftData,
    SpecDriftHeader,
    StringData,
    TipTrackHeader,
    TipTrackData,
    PrmHeader,
    PiezoSensitivity,
    FrequencySweepData,
    ScanProcessorInfo,
    PllInfo,
    ChannelDriveInfo,
    LockinInfo,
    PiControllerInfo,
    LowpassFilterInfo,
)

# TypeAlias
FileHeaderObject = Union[Prm, PrmHeader]

# TypeAlias
PageHeaderObject = Union[
    ApiInfo,
    ImageDriftHeader,
    ImageDriftData,
    SpecDriftData,
    SpecDriftHeader,
    StringData,
    TipTrackHeader,
    TipTrackData,
    PiezoSensitivity,
    FrequencySweepData,
    ScanProcessorInfo,
    PllInfo,
    ChannelDriveInfo,
    LockinInfo,
    PiControllerInfo,
    LowpassFilterInfo,
]

# TypeAlias
# PageObject = Union[
#     PageData,
#     Thumbnail,
#      ThumbnailHeader,
# ]


class RhkPageDataType(Enum):
    RHK_DATA_IMAGE = 0
    RHK_DATA_LINE = 1
    RHK_DATA_XY_DATA = 2
    RHK_DATA_ANNOTATED_LINE = 3
    RHK_DATA_TEXT = 4
    RHK_DATA_ANNOTATED_TEXT = 5
    RHK_DATA_SEQUENTIAL = 6
    RHK_DATA_MOVIE = 7


class RhkObjectType(Enum):
    RHK_OBJECT_UNDEFINED = 0
    RHK_OBJECT_PAGE_INDEX_HEADER = 1
    RHK_OBJECT_PAGE_INDEX_ARRAY = 2
    RHK_OBJECT_PAGE_HEADER = 3
    RHK_OBJECT_PAGE_DATA = 4
    RHK_OBJECT_IMAGE_DRIFT_HEADER = 5
    RHK_OBJECT_IMAGE_DRIFT = 6
    RHK_OBJECT_SPEC_DRIFT_HEADER = 7
    RHK_OBJECT_SPEC_DRIFT_DATA = 8
    RHK_OBJECT_COLOR_INFO = 9
    RHK_OBJECT_STRING_DATA = 10
    RHK_OBJECT_TIP_TRACK_HEADER = 11
    RHK_OBJECT_TIP_TRACK_DATA = 12
    RHK_OBJECT_PRM = 13
    RHK_OBJECT_THUMBNAIL = 14
    RHK_OBJECT_PRM_HEADER = 15
    RHK_OBJECT_THUMBNAIL_HEADER = 16
    RHK_OBJECT_API_INFO = 17
    RHK_OBJECT_HISTORY_INFO = 18
    RHK_OBJECT_PIEZO_SENSITIVITY = 19
    RHK_OBJECT_FREQUENCY_SWEEP_DATA = 20
    RHK_OBJECT_SCAN_PROCESSOR_INFO = 21
    RHK_OBJECT_PLL_INFO = 22
    RHK_OBJECT_CH1_DRIVE_INFO = 23
    RHK_OBJECT_CH2_DRIVE_INFO = 24
    RHK_OBJECT_LOCKIN0_INFO = 25
    RHK_OBJECT_LOCKIN1_INFO = 26
    RHK_OBJECT_ZPI_INFO = 27
    RHK_OBJECT_KPI_INFO = 28
    RHK_OBJECT_AUX_PI_INFO = 29
    RHK_OBJECT_LOWPASS_FILTER0_INFO = 30
    RHK_OBJECT_LOWPASS_FILTER1_INFO = 31


class RhkPageSourceType(Enum):
    RHK_SOURCE_RAW = 0
    RHK_SOURCE_PROCESSED = 1
    RHK_SOURCE_CALCULATED = 2
    RHK_SOURCE_IMPORTED = 3


class RhkPageType(Enum):
    RHK_PAGE_UNDEFINED = 0
    RHK_PAGE_TOPOGRAPHIC = 1
    RHK_PAGE_CURRENT = 2
    RHK_PAGE_AUX = 3
    RHK_PAGE_FORCE = 4
    RHK_PAGE_SIGNAL = 5
    RHK_PAGE_FFT_TRANSFORM = 6
    RHK_PAGE_NOISE_POWER_SPECTRUM = 7
    RHK_PAGE_LINE_TEST = 8
    RHK_PAGE_OSCILLOSCOPE = 9
    RHK_PAGE_IV_SPECTRA = 10
    RHK_PAGE_IV_4x4 = 11
    RHK_PAGE_IV_8x8 = 12
    RHK_PAGE_IV_16x16 = 13
    RHK_PAGE_IV_32x32 = 14
    RHK_PAGE_IV_CENTER = 15
    RHK_PAGE_INTERACTIVE_SPECTRA = 16
    RHK_PAGE_AUTOCORRELATION = 17
    RHK_PAGE_IZ_SPECTRA = 18
    RHK_PAGE_4_GAIN_TOPOGRAPHY = 19
    RHK_PAGE_8_GAIN_TOPOGRAPHY = 20
    RHK_PAGE_4_GAIN_CURRENT = 21
    RHK_PAGE_8_GAIN_CURRENT = 22
    RHK_PAGE_IV_64x64 = 23
    RHK_PAGE_AUTOCORRELATION_SPECTRUM = 24
    RHK_PAGE_COUNTER = 25
    RHK_PAGE_MULTICHANNEL_ANALYSER = 26
    RHK_PAGE_AFM_100 = 27
    RHK_PAGE_CITS = 28
    RHK_PAGE_GPIB = 29
    RHK_PAGE_VIDEO_CHANNEL = 30
    RHK_PAGE_IMAGE_OUT_SPECTRA = 31
    RHK_PAGE_I_DATALOG = 32
    RHK_PAGE_I_ECSET = 33
    RHK_PAGE_I_ECDATA = 34
    RHK_PAGE_I_DSP_AD = 35
    RHK_PAGE_DISCRETE_SPECTROSCOPY_PP = 36
    RHK_PAGE_IMAGE_DISCRETE_SPECTROSCOPY = 37
    RHK_PAGE_RAMP_SPECTROSCOPY_RP = 38
    RHK_PAGE_DISCRETE_SPECTROSCOPY_RP = 39


class RhkLineType(Enum):
    RHK_LINE_NOT_A_LINE = 0
    RHK_LINE_HISTOGRAM = 1
    RHK_LINE_CROSS_SECTION = 2
    RHK_LINE_LINE_TEST = 3
    RHK_LINE_OSCILLOSCOPE = 4
    RHK_LINE_RESERVED = 5
    RHK_LINE_NOISE_POWER_SPECTRUM = 6
    RHK_LINE_IV_SPECTRUM = 7
    RHK_LINE_IZ_SPECTRUM = 8
    RHK_LINE_IMAGE_X_AVERAGE = 9
    RHK_LINE_IMAGE_Y_AVERAGE = 10
    RHK_LINE_NOISE_AUTOCORRELATION_SPECTRUM = 11
    RHK_LINE_MULTICHANNEL_ANALYSER_DATA = 12
    RHK_LINE_RENORMALIZED_IV = 13
    RHK_LINE_IMAGE_HISTOGRAM_SPECTRA = 14
    RHK_LINE_IMAGE_CROSS_SECTION = 15
    RHK_LINE_IMAGE_AVERAGE = 16
    RHK_LINE_IMAGE_CROSS_SECTION_G = 17
    RHK_LINE_IMAGE_OUT_SPECTRA = 18
    RHK_LINE_DATALOG_SPECTRUM = 19
    RHK_LINE_GXY = 20
    RHK_LINE_ELECTROCHEMISTRY = 21
    RHK_LINE_DISCRETE_SPECTROSCOPY = 22
    RHK_LINE_DATA_LOGGER = 23
    RHK_LINE_TIME_SPECTROSCOPY = 24
    RHK_LINE_ZOOM_FFT = 25
    RHK_LINE_FREQUENCY_SWEEP = 26
    RHK_LINE_PHASE_ROTATE = 27
    RHK_LINE_FIBER_SWEEP = 28


class RhkImageType(Enum):
    RHK_IMAGE_NORMAL = 0
    RHK_IMAGE_AUTOCORRELATED = 1


class RhkScanType(Enum):
    RHK_SCAN_RIGHT = 0
    RHK_SCAN_LEFT = 1
    RHK_SCAN_UP = 2
    RHK_SCAN_DOWN = 3

    def direction(self) -> str:
        if self == RhkScanType.RHK_SCAN_RIGHT:
            return "right"
        elif self == RhkScanType.RHK_SCAN_LEFT:
            return "left"
        elif self == RhkScanType.RHK_SCAN_UP:
            return "up"
        elif self == RhkScanType.RHK_SCAN_DOWN:
            return "down"
        else:
            return "unknown"


@dataclass
class Sm4Object:
    obj_type: RhkObjectType
    offset: int
    size: int

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> Sm4Object:
        object_type_id = RhkObjectType(cursor.read_u32_le())
        offset = cursor.read_u32_le()
        size = cursor.read_u32_le()
        return cls(object_type_id, offset, size)


@dataclass
class Sm4FileHeader:
    """Its object_list contains PAGE_INDEX_ARRAY, PRM, PRM_HEADER"""

    size: int
    signature: str
    page_count: int
    object_list_count: int
    object_field_size: int
    object_list: List[Sm4Object]
    page_index_header: Sm4PageIndexHeader = field(init=False)
    prm_header: PrmHeader = field(init=False)
    prm: Prm = field(init=False)

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> Sm4FileHeader:
        size = cursor.read_u16_le()
        signature = cursor.read_string(36)
        page_count = cursor.read_u32_le()
        object_list_count = cursor.read_u32_le()
        object_field_size = cursor.read_u32_le()

        _ = cursor.read_u32_le()
        _ = cursor.read_u32_le()

        object_list: List[Sm4Object] = []
        for _ in range(object_list_count):
            obj = Sm4Object.from_buffer(cursor)
            """ print(obj) """
            object_list.append(obj)

        return cls(
            size,
            signature,
            page_count,
            object_list_count,
            object_field_size,
            object_list,
        )

    def read_objects(self, cursor: Cursor) -> None:
        """Read PRM header first"""

        self.read_prm_header(cursor)
        if self.prm_header is None:
            raise BufferError("No PRM header in file header")

        for obj in self.object_list:
            cursor.set_position(obj.offset)
            if obj.obj_type == RhkObjectType.RHK_OBJECT_PAGE_INDEX_HEADER:
                self.page_index_header = Sm4PageIndexHeader.from_buffer(
                    cursor, obj.offset
                )

            if obj.obj_type == RhkObjectType.RHK_OBJECT_PRM:
                self.prm = Prm.from_buffer(
                    cursor,
                    self.prm_header.prm_compression_flag,
                    self.prm_header.prm_data_size,
                    self.prm_header.prm_compression_size,
                )

    def read_prm_header(self, cursor: Cursor) -> None:
        for obj in self.object_list:
            if obj.obj_type == RhkObjectType.RHK_OBJECT_PRM_HEADER:
                cursor.set_position(obj.offset)
                self.prm_header = PrmHeader.from_buffer(cursor)


@dataclass
class Sm4PageIndexHeader:
    offset: int
    page_count: int
    object_list_count: int
    object_list: List[Sm4Object]

    @classmethod
    def from_buffer(cls, cursor: Cursor, offset: int) -> Sm4PageIndexHeader:
        page_count = cursor.read_u32_le()
        object_list_count = cursor.read_u32_le()
        _reserved_1 = cursor.read_u32_le()  # pyright: ignore
        _reserved_2 = cursor.read_u32_le()  # pyright: ignore

        object_list = []
        for _ in range(object_list_count):
            object_list.append(Sm4Object.from_buffer(cursor))

        return Sm4PageIndexHeader(
            offset, page_count, object_list_count, object_list
        )

    def page_index_array_offset(self) -> Optional[int]:
        for obj in self.object_list:
            if obj.obj_type == RhkObjectType.RHK_OBJECT_PAGE_INDEX_ARRAY:
                return obj.offset
        else:
            raise BufferError("No Page Index Array found")


@dataclass
class Sm4PageHeaderSequential:
    data_type: int
    data_length: int
    param_count: int
    object_list_count: int
    data_info_size: int
    data_info_string_count: int
    object_list: List[Sm4Object]

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> Sm4PageHeaderSequential:
        data_type = cursor.read_u32_le()
        data_length = cursor.read_u32_le()
        param_count = cursor.read_u32_le()
        object_list_count = cursor.read_u32_le()
        data_info_size = cursor.read_u32_le()
        data_info_string_count = cursor.read_u32_le()

        object_list: List[Sm4Object] = []
        for _ in range(object_list_count):
            object_list.append(Sm4Object.from_buffer(cursor))

        sequential_param_gain: List[float] = []
        sequential_param_label: List[str] = []
        sequential_param_unit: List[str] = []
        for _ in range(param_count):
            sequential_param_gain.append(cursor.read_f32_le())
            sequential_param_label.append(cursor.read_sm4_string())
            sequential_param_unit.append(cursor.read_sm4_string())

        return Sm4PageHeaderSequential(
            data_type,
            data_length,
            param_count,
            object_list_count,
            data_info_size,
            data_info_string_count,
            object_list,
        )


@dataclass
class Sm4PageHeaderDefault:
    string_count: int
    page_type: RhkPageType
    data_sub_source: int
    line_type: RhkLineType
    x_corner: int
    y_corner: int
    x_size: int
    y_size: int
    image_type: RhkImageType
    scan_type: RhkScanType
    group_id: int
    page_data_size: int
    min_z_value: int
    max_z_value: int
    x_scale: float
    y_scale: float
    z_scale: float
    xy_scale: float
    x_offset: float
    y_offset: float
    z_offset: float
    period: float
    bias: float
    current: float
    angle: float
    color_info_count: int
    grid_x_size: int
    grid_y_size: int
    object_list_count: int
    _32_bit_data_flag: int
    object_list: List[Sm4Object]
    page_header_objects: List[PageHeaderObject] = field(default_factory=list)

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> Sm4PageHeaderDefault:
        _ = cursor.read_u16_le()
        string_count = cursor.read_u16_le()
        page_type = RhkPageType(cursor.read_u32_le())
        data_sub_source = cursor.read_u32_le()

        line_type = RhkLineType(cursor.read_u32_le())

        x_corner = cursor.read_u32_le()
        y_corner = cursor.read_u32_le()
        x_size = cursor.read_u32_le()
        y_size = cursor.read_u32_le()

        image_type = RhkImageType(cursor.read_u32_le())

        scan_type = RhkScanType(cursor.read_u32_le())

        group_id = cursor.read_u32_le()
        page_data_size = cursor.read_u32_le()

        min_z_value = cursor.read_u32_le()
        max_z_value = cursor.read_u32_le()

        x_scale = cursor.read_f32_le()
        y_scale = cursor.read_f32_le()
        z_scale = cursor.read_f32_le()
        xy_scale = cursor.read_f32_le()
        x_offset = cursor.read_f32_le()
        y_offset = cursor.read_f32_le()
        z_offset = cursor.read_f32_le()
        period = cursor.read_f32_le()
        bias = cursor.read_f32_le()
        current = cursor.read_f32_le()
        angle = cursor.read_f32_le()

        color_info_count = cursor.read_u32_le()
        grid_x_size = cursor.read_u32_le()
        grid_y_size = cursor.read_u32_le()

        object_list_count = cursor.read_u32_le()
        _32_bit_data_flag = cursor.read_u8_le()

        # reserved
        cursor.skip(63)

        object_list: List[Sm4Object] = []
        for _ in range(object_list_count):
            obj = Sm4Object.from_buffer(cursor)
            """ print("PageHeaderDefault : ", obj) """
            object_list.append(obj)

        return Sm4PageHeaderDefault(
            string_count,
            page_type,
            data_sub_source,
            line_type,
            x_corner,
            y_corner,
            x_size,
            y_size,
            image_type,
            scan_type,
            group_id,
            page_data_size,
            min_z_value,
            max_z_value,
            x_scale,
            y_scale,
            z_scale,
            xy_scale,
            x_offset,
            y_offset,
            z_offset,
            period,
            bias,
            current,
            angle,
            color_info_count,
            grid_x_size,
            grid_y_size,
            object_list_count,
            _32_bit_data_flag,
            object_list,
        )

    def read_data(self, cursor: Cursor) -> None:
        tiptrack_info_count = None
        for obj in self.object_list:
            if obj.offset != 0 and obj.size != 0:
                cursor.set_position(obj.offset)

                if obj.obj_type == RhkObjectType.RHK_OBJECT_IMAGE_DRIFT_HEADER:
                    self.page_header_objects.append(
                        ImageDriftHeader.from_buffer(cursor)
                    )

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_IMAGE_DRIFT:
                    self.page_header_objects.append(
                        ImageDriftData.from_buffer(cursor)
                    )

                elif (
                    obj.obj_type == RhkObjectType.RHK_OBJECT_SPEC_DRIFT_HEADER
                ):
                    self.page_header_objects.append(
                        SpecDriftHeader.from_buffer(cursor)
                    )

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_SPEC_DRIFT_DATA:
                    self.page_header_objects.append(
                        SpecDriftData.from_buffer(cursor, self.y_size)
                    )

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_COLOR_INFO:
                    pass

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_STRING_DATA:
                    self.page_header_objects.append(
                        StringData.from_buffer(cursor, self.string_count)
                    )

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_TIP_TRACK_HEADER:
                    read_obj = TipTrackHeader.from_buffer(cursor)
                    tiptrack_info_count = read_obj.tiptrack_tiptrack_info_count
                    self.page_header_objects.append(read_obj)

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_TIP_TRACK_DATA:
                    if tiptrack_info_count is not None:
                        self.page_header_objects.append(
                            TipTrackData.from_buffer(
                                cursor, tiptrack_info_count
                            )
                        )
                    else:
                        raise ValueError("tiptrack_info_count not found")

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_THUMBNAIL:
                    pass

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_THUMBNAIL_HEADER:
                    pass

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_API_INFO:
                    self.page_header_objects.append(
                        ApiInfo.from_buffer(cursor)
                    )

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_HISTORY_INFO:
                    pass

                elif (
                    obj.obj_type == RhkObjectType.RHK_OBJECT_PIEZO_SENSITIVITY
                ):
                    self.page_header_objects.append(
                        PiezoSensitivity.from_buffer(cursor)
                    )

                elif (
                    obj.obj_type
                    == RhkObjectType.RHK_OBJECT_FREQUENCY_SWEEP_DATA
                ):
                    self.page_header_objects.append(
                        FrequencySweepData.from_buffer(cursor)
                    )

                elif (
                    obj.obj_type
                    == RhkObjectType.RHK_OBJECT_SCAN_PROCESSOR_INFO
                ):
                    self.page_header_objects.append(
                        ScanProcessorInfo.from_buffer(cursor)
                    )

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_PLL_INFO:
                    self.page_header_objects.append(
                        PllInfo.from_buffer(cursor)
                    )

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_CH1_DRIVE_INFO:
                    self.page_header_objects.append(
                        ChannelDriveInfo.from_buffer(cursor)
                    )

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_CH2_DRIVE_INFO:
                    self.page_header_objects.append(
                        ChannelDriveInfo.from_buffer(cursor)
                    )

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_LOCKIN0_INFO:
                    self.page_header_objects.append(
                        LockinInfo.from_buffer(cursor)
                    )

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_LOCKIN1_INFO:
                    self.page_header_objects.append(
                        LockinInfo.from_buffer(cursor)
                    )

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_ZPI_INFO:
                    self.page_header_objects.append(
                        PiControllerInfo.from_buffer(cursor)
                    )

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_KPI_INFO:
                    self.page_header_objects.append(
                        PiControllerInfo.from_buffer(cursor)
                    )

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_AUX_PI_INFO:
                    self.page_header_objects.append(
                        PiControllerInfo.from_buffer(cursor)
                    )

                elif (
                    obj.obj_type
                    == RhkObjectType.RHK_OBJECT_LOWPASS_FILTER0_INFO
                ):
                    self.page_header_objects.append(
                        LowpassFilterInfo.from_buffer(cursor)
                    )

                elif (
                    obj.obj_type
                    == RhkObjectType.RHK_OBJECT_LOWPASS_FILTER1_INFO
                ):
                    self.page_header_objects.append(
                        LowpassFilterInfo.from_buffer(cursor)
                    )


# TypeAlias
Sm4PageHeader = Union[Sm4PageHeaderSequential, Sm4PageHeaderDefault]


@dataclass
class Sm4Page:
    """Its object_list contains PAGE_HEADER, PAGE_DATA, THUMBNAIL and
    THUMBNAIL_HEADER
    """

    header: Sm4PageHeader = field(init=False)
    data: PageData = field(init=False)
    label: str = field(init=False)
    page_id: int
    page_data_type: RhkPageDataType
    page_source_type: RhkPageSourceType
    object_list_count: int
    minor_version: int
    object_list: List[Sm4Object]
    page_objects: List[PageHeaderObject] = field(default_factory=list)

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> Sm4Page:
        page_id = cursor.read_u16_le()
        cursor.skip(14)
        page_data_type = RhkPageDataType(cursor.read_u32_le())
        page_source_type = RhkPageSourceType(cursor.read_u32_le())
        object_list_count = cursor.read_u32_le()
        minor_version = cursor.read_u32_le()

        object_list = []
        for _ in range(object_list_count):
            object_list.append(Sm4Object.from_buffer(cursor))

        return cls(
            page_id,
            page_data_type,
            page_source_type,
            minor_version,
            object_list_count,
            object_list,
        )

    def add_header(self, header: Sm4PageHeader) -> None:
        self.header = header

    def read_data(self, cursor: Cursor) -> None:
        for obj in self.object_list:
            if (
                obj.offset != 0
                and obj.size != 0
                and isinstance(self.header, Sm4PageHeaderDefault)
            ):
                cursor.set_position(obj.offset)

                if obj.obj_type == RhkObjectType.RHK_OBJECT_PAGE_DATA:
                    self.data = PageData.from_buffer(
                        cursor,
                        obj.size,
                        self.header.z_scale,
                        self.header.z_offset,
                    )

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_THUMBNAIL:
                    pass

                elif obj.obj_type == RhkObjectType.RHK_OBJECT_THUMBNAIL_HEADER:
                    pass

    def read_label(self) -> None:
        if type(self.header) == Sm4PageHeaderDefault:
            for obj in self.header.page_header_objects:
                if type(obj) == StringData:
                    self.label = obj.label


class Sm4FileAll:
    def __init__(self, filepath: str):
        self.filepath = filepath
        with open(filepath, "rb") as f:
            self._read_sm4_file(f)

    def _read_sm4_file(self, f: BufferedReader) -> None:
        cursor = Cursor(f)
        self.file_header = Sm4FileHeader.from_buffer(cursor)
        self.file_header.read_objects(cursor)
        page_index_header = self.file_header.page_index_header

        page_index_array_offset = page_index_header.page_index_array_offset()
        assert page_index_array_offset
        cursor.set_position(page_index_array_offset)

        self.pages: List[Sm4Page] = []
        for _ in range(page_index_header.page_count):
            page = Sm4Page.from_buffer(cursor)
            self.pages.append(page)

        offset = None
        for page in self.pages:
            for obj in page.object_list:
                if obj.obj_type == RhkObjectType.RHK_OBJECT_PAGE_HEADER:
                    offset = obj.offset
                    break
            if offset is None:
                raise BufferError(f"No page header in page")
            cursor.set_position(offset)

            if page.page_data_type == RhkPageDataType.RHK_DATA_SEQUENTIAL:
                page_header: Sm4PageHeader = (
                    Sm4PageHeaderSequential.from_buffer(cursor)
                )

            else:
                page_header = Sm4PageHeaderDefault.from_buffer(cursor)
                page_header.read_data(cursor)

            page.add_header(page_header)
            page.read_label()
            page.read_data(cursor)

        self._arrange_data()

    def _arrange_data(self) -> None:
        for page in self.pages:
            if type(page.header) == Sm4PageHeaderDefault:
                page.data.data = page.data.data.reshape(
                    page.header.y_size, page.header.x_size
                )

                if page.page_data_type == RhkPageDataType.RHK_DATA_IMAGE:
                    if page.header.x_scale < 0:
                        page.data.data = np.flip(page.data.data, axis=1)
                    if page.header.y_scale > 0:
                        page.data.data = np.flip(page.data.data, axis=0)

                elif page.page_data_type == RhkPageDataType.RHK_DATA_LINE:
                    x_values = (
                        np.arange(page.header.x_size) * page.header.x_scale
                        + page.header.x_offset
                    )
                    y_values_arr = page.data.data.transpose()
                    page.data.data = np.column_stack((x_values, y_values_arr))



from __future__ import annotations
from typing import List
from enum import Enum
from dataclasses import dataclass
import zlib

import numpy as np
from numpy._typing import NDArray

from .cursor import Cursor


class RhkDriftOptionType(Enum):
    RHK_DRIFT_DISABLED = 0
    RHK_DRIFT_EACH_SPECTRA = 1
    RHK_DRIFT_EACH_LOCATION = 2


@dataclass
class PageData:
    data: NDArray[np.float32]

    @classmethod
    def from_buffer(
        cls, cursor: Cursor, size: int, z_scale: float, z_offset: float
    ) -> PageData:
        # cursor.set_position(offset)

        raw_data = np.frombuffer(cursor.read(size), dtype=np.int32)
        data: NDArray[np.float32] = raw_data * z_scale + z_offset
        return cls(data)


@dataclass
class ImageDriftHeader:
    imagedrift_filetime: int
    imagedrift_drift_option_type: RhkDriftOptionType

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> ImageDriftHeader:
        imagedrift_filetime = cursor.read_u64_le()
        imagedrift_drift_option_type = RhkDriftOptionType(cursor.read_u32_le())

        return cls(imagedrift_filetime, imagedrift_drift_option_type)


@dataclass
class ImageDriftData:
    imagedrift_time: int
    imagedrift_dx: int
    imagedrift_dy: int
    imagedrift_cumulative_x: int
    imagedrift_cumulative_y: int
    imagedrift_vector_x: int
    imagedrift_vector_y: int

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> ImageDriftData:
        imagedrift_time = cursor.read_u32_le()
        imagedrift_dx = cursor.read_u32_le()
        imagedrift_dy = cursor.read_u32_le()
        imagedrift_cumulative_x = cursor.read_u32_le()
        imagedrift_cumulative_y = cursor.read_u32_le()
        imagedrift_vector_x = cursor.read_u32_le()
        imagedrift_vector_y = cursor.read_u32_le()

        return cls(
            imagedrift_time,
            imagedrift_dx,
            imagedrift_dy,
            imagedrift_cumulative_x,
            imagedrift_cumulative_y,
            imagedrift_vector_x,
            imagedrift_vector_y,
        )


@dataclass
class SpecDriftHeader:
    specdrift_filetime: int
    specdrift_drift_option_type: int
    specdrift_drift_option_type_name: str
    specdrift_channel: str

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> SpecDriftHeader:
        # unix epoch
        specdrift_filetime = cursor.read_u64_le()
        specdrift_drift_option_type = cursor.read_u32_le()
        if specdrift_drift_option_type == 0:
            specdrift_drift_option_type_name = "RHK_DRIFT_DISABLED"
        elif specdrift_drift_option_type == 1:
            specdrift_drift_option_type_name = "RHK_DRIFT_EACH_SPECTRA"
        elif specdrift_drift_option_type == 1:
            specdrift_drift_option_type_name = "RHK_DRIFT_EACH_LOCATION"
        else:
            specdrift_drift_option_type_name = "RHK_DRIFT_UNKNOWN"

        _ = cursor.read_u32_le()
        specdrift_channel = cursor.read_sm4_string()

        return cls(
            specdrift_filetime,
            specdrift_drift_option_type,
            specdrift_drift_option_type_name,
            specdrift_channel,
        )


@dataclass
class SpecDriftData:
    specdrift_time: List[float]
    specdrift_x_coord: List[float]
    specdrift_y_coord: List[float]
    specdrift_dx: List[float]
    specdrift_dy: List[float]
    specdrift_cumulative_x: List[float]
    specdrift_cumulative_y: List[float]

    @classmethod
    def from_buffer(cls, cursor: Cursor, y_size: int) -> SpecDriftData:
        specdrift_time: List[float] = []
        specdrift_x_coord: List[float] = []
        specdrift_y_coord: List[float] = []
        specdrift_dx: List[float] = []
        specdrift_dy: List[float] = []
        specdrift_cumulative_x: List[float] = []
        specdrift_cumulative_y: List[float] = []

        for _ in range(y_size):
            specdrift_time.append(cursor.read_f32_le())
            specdrift_x_coord.append(cursor.read_f32_le())
            specdrift_y_coord.append(cursor.read_f32_le())
            specdrift_dx.append(cursor.read_f32_le())
            specdrift_dy.append(cursor.read_f32_le())
            specdrift_cumulative_x.append(cursor.read_f32_le())
            specdrift_cumulative_y.append(cursor.read_f32_le())

        return cls(
            specdrift_time,
            specdrift_x_coord,
            specdrift_y_coord,
            specdrift_dx,
            specdrift_dy,
            specdrift_cumulative_x,
            specdrift_cumulative_y,
        )


@dataclass
class StringData:
    label: str
    system_text: str
    session_text: str
    user_text: str
    filename: str
    date: str
    time: str
    x_units: str
    y_units: str
    z_units: str
    x_label: str
    y_label: str
    status_channel_text: str
    completed_line_count: str
    oversampling_count: str
    sliced_voltage: str
    pll_pro_status: str
    setpoint_unit: str
    channel_list: str

    @classmethod
    def from_buffer(cls, cursor: Cursor, count: int) -> StringData:
        strings = [cursor.read_sm4_string() for _ in range(count)]

        label = strings[0]
        system_text = strings[1]
        session_text = strings[2]
        user_text = strings[3]
        filename = strings[4]
        date = strings[5]
        time = strings[6]
        x_units = strings[7]
        y_units = strings[8]
        z_units = strings[9]
        x_label = strings[10]
        y_label = strings[11]
        status_channel_text = strings[12]
        completed_line_count = strings[13]
        try:
            oversampling_count = strings[14]
            sliced_voltage = strings[15]
            pll_pro_status = strings[16]
            setpoint_unit = strings[17]
            channel_list = strings[18]
        except IndexError:
            oversampling_count = ""
            sliced_voltage = ""
            pll_pro_status = ""
            setpoint_unit = ""
            channel_list = ""

        return cls(
            label,
            system_text,
            session_text,
            user_text,
            filename,
            date,
            time,
            x_units,
            y_units,
            z_units,
            x_label,
            y_label,
            status_channel_text,
            completed_line_count,
            oversampling_count,
            sliced_voltage,
            pll_pro_status,
            setpoint_unit,
            channel_list,
        )


@dataclass
class TipTrackHeader:
    tiptrack_filetime: int
    tiptrack_feature_height: float
    tiptrack_feature_width: float
    tiptrack_time_constant: float
    tiptrack_cycle_rate: float
    tiptrack_phase_lag: float
    tiptrack_tiptrack_info_count: int
    tiptrack_channel: str

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> TipTrackHeader:
        # unix epoch
        tiptrack_filetime = cursor.read_u64_le()
        tiptrack_feature_height = cursor.read_f32_le()
        tiptrack_feature_width = cursor.read_f32_le()
        tiptrack_time_constant = cursor.read_f32_le()
        tiptrack_cycle_rate = cursor.read_f32_le()
        tiptrack_phase_lag = cursor.read_f32_le()
        _ = cursor.read_u32_le()
        tiptrack_tiptrack_info_count = cursor.read_u32_le()
        tiptrack_channel = cursor.read_sm4_string()

        return cls(
            tiptrack_filetime,
            tiptrack_feature_height,
            tiptrack_feature_width,
            tiptrack_time_constant,
            tiptrack_cycle_rate,
            tiptrack_phase_lag,
            tiptrack_tiptrack_info_count,
            tiptrack_channel,
        )


@dataclass
class TipTrackData:
    tiptrack_cumulative_time: List[float]
    tiptrack_time: List[float]
    tiptrack_dx: List[float]
    tiptrack_dy: List[float]

    @classmethod
    def from_buffer(
        cls, cursor: Cursor, tiptrack_tiptrack_info_count: int
    ) -> TipTrackData:
        tiptrack_cumulative_time: List[float] = []
        tiptrack_time: List[float] = []
        tiptrack_dx: List[float] = []
        tiptrack_dy: List[float] = []

        for _ in range(tiptrack_tiptrack_info_count):
            tiptrack_cumulative_time.append(cursor.read_f32_le())
            tiptrack_time.append(cursor.read_f32_le())
            tiptrack_dx.append(cursor.read_f32_le())
            tiptrack_dy.append(cursor.read_f32_le())

        return cls(
            tiptrack_cumulative_time,
            tiptrack_time,
            tiptrack_dx,
            tiptrack_dy,
        )


@dataclass
class Prm:
    prm_data: str

    @classmethod
    def from_buffer(
        cls,
        cursor: Cursor,
        prm_compression_flag: int,
        prm_data_size: int,
        prm_compression_size: int,
    ) -> Prm:
        if prm_compression_flag == 0:
            prm_data_raw = cursor.read(prm_data_size)
        else:
            prm_data_raw = zlib.decompress(
                cursor.read(prm_compression_size),
                wbits=0,
                bufsize=prm_data_size,
            )

        prm_data = prm_data_raw.decode("CP437")
        return cls(prm_data)


@dataclass
class PrmHeader:
    prm_compression_flag: int
    prm_data_size: int
    prm_compression_size: int

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> PrmHeader:
        prm_compression_flag = cursor.read_u32_le()
        prm_data_size = cursor.read_u32_le()
        prm_compression_size = cursor.read_u32_le()

        # read the actual PRM data

        return cls(
            prm_compression_flag,
            prm_data_size,
            prm_compression_size,
        )


@dataclass
class ApiInfo:
    voltage_high: float
    voltage_low: float
    gain: float
    api_offset: float
    ramp_mode: int
    ramp_type: int
    step: int
    image_count: int
    dac: int
    mux: int
    bias: int
    units: str

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> ApiInfo:
        voltage_high = cursor.read_f32_le()
        voltage_low = cursor.read_f32_le()
        gain = cursor.read_f32_le()
        api_offset = cursor.read_f32_le()

        ramp_mode = cursor.read_u32_le()
        ramp_type = cursor.read_u32_le()
        step = cursor.read_u32_le()
        image_count = cursor.read_u32_le()
        dac = cursor.read_u32_le()
        mux = cursor.read_u32_le()
        bias = cursor.read_u32_le()

        _ = cursor.read_u32_le()
        units = cursor.read_sm4_string()

        return cls(
            voltage_high,
            voltage_low,
            gain,
            api_offset,
            ramp_mode,
            ramp_type,
            step,
            image_count,
            dac,
            mux,
            bias,
            units,
        )


@dataclass
class PiezoSensitivity:
    tube_x: float
    tube_y: float
    tube_z: float
    tube_z_offset: float
    scan_x: float
    scan_y: float
    scan_z: float
    actuator: float
    tube_x_unit: str
    tube_y_unit: str
    tube_z_unit: str
    tube_z_unit_offset: str
    scan_x_unit: str
    scan_y_unit: str
    scan_z_unit: str
    actuator_unit: str
    tube_calibration: str
    scan_calibration: str
    actuator_calibration: str

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> PiezoSensitivity:
        tube_x = cursor.read_f64_le()
        tube_y = cursor.read_f64_le()
        tube_z = cursor.read_f64_le()
        tube_z_offset = cursor.read_f64_le()
        scan_x = cursor.read_f64_le()
        scan_y = cursor.read_f64_le()
        scan_z = cursor.read_f64_le()
        actuator = cursor.read_f64_le()

        _ = cursor.read_u32_le()

        tube_x_unit = cursor.read_sm4_string()
        tube_y_unit = cursor.read_sm4_string()
        tube_z_unit = cursor.read_sm4_string()
        tube_z_unit_offset = cursor.read_sm4_string()
        scan_x_unit = cursor.read_sm4_string()
        scan_y_unit = cursor.read_sm4_string()
        scan_z_unit = cursor.read_sm4_string()
        actuator_unit = cursor.read_sm4_string()
        tube_calibration = cursor.read_sm4_string()
        scan_calibration = cursor.read_sm4_string()
        actuator_calibration = cursor.read_sm4_string()

        return cls(
            tube_x,
            tube_y,
            tube_z,
            tube_z_offset,
            scan_x,
            scan_y,
            scan_z,
            actuator,
            tube_x_unit,
            tube_y_unit,
            tube_z_unit,
            tube_z_unit_offset,
            scan_x_unit,
            scan_y_unit,
            scan_z_unit,
            actuator_unit,
            tube_calibration,
            scan_calibration,
            actuator_calibration,
        )


@dataclass
class FrequencySweepData:
    psd_total_signal: float
    peak_frequency: float
    peak_amplitude: float
    drive_aplitude: float
    signal_to_drive_ratio: float
    q_factor: float
    total_signal_unit: str
    peak_frequency_unit: str
    peak_amplitude_unit: str
    drive_amplitude_unit: str
    signal_to_drive_ratio_unit: str
    q_factor_unit: str

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> FrequencySweepData:
        psd_total_signal = cursor.read_f64_le()
        peak_frequency = cursor.read_f64_le()
        peak_amplitude = cursor.read_f64_le()
        drive_aplitude = cursor.read_f64_le()
        signal_to_drive_ratio = cursor.read_f64_le()
        q_factor = cursor.read_f64_le()

        _ = cursor.read_u32_le()

        total_signal_unit = cursor.read_sm4_string()
        peak_frequency_unit = cursor.read_sm4_string()
        peak_amplitude_unit = cursor.read_sm4_string()
        drive_amplitude_unit = cursor.read_sm4_string()
        signal_to_drive_ratio_unit = cursor.read_sm4_string()
        q_factor_unit = cursor.read_sm4_string()

        return cls(
            psd_total_signal,
            peak_frequency,
            peak_amplitude,
            drive_aplitude,
            signal_to_drive_ratio,
            q_factor,
            total_signal_unit,
            peak_frequency_unit,
            peak_amplitude_unit,
            drive_amplitude_unit,
            signal_to_drive_ratio_unit,
            q_factor_unit,
        )


@dataclass
class ScanProcessorInfo:
    x_slope_compensation: float
    y_slope_compensation: float
    x_slope_compensation_unit: str
    y_slope_compensation_unit: str

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> ScanProcessorInfo:
        x_slope_compensation = cursor.read_f64_le()
        y_slope_compensation = cursor.read_f64_le()
        _ = cursor.read_u32_le()
        x_slope_compensation_unit = cursor.read_sm4_string()
        y_slope_compensation_unit = cursor.read_sm4_string()

        return cls(
            x_slope_compensation,
            y_slope_compensation,
            x_slope_compensation_unit,
            y_slope_compensation_unit,
        )


@dataclass
class PllInfo:
    amplitude_control: int
    drive_amplitude: float
    drive_ref_frequency: float
    lockin_freq_offset: float
    lockin_harmonic_factor: float
    lockin_phase_offset: float
    pi_gain: float
    pi_int_cutoff_frequency: float
    pi_lower_bound: float
    pi_upper_bound: float
    diss_pi_gain: float
    diss_pi_int_cutoff_frequency: float
    diss_pi_lower_bound: float
    diss_pi_upper_bound: float
    lockin_filter_cutoff_frequency: str
    drive_amplitude_unit: str
    drive_ref_frequency_unit: str
    lockin_freq_offset_unit: str
    lockin_harmonic_factor_unit: str
    lockin_phase_offset_unit: str
    pi_gain_unit: str
    pi_int_cutoff_frequency_unit: str
    pi_lower_bound_unit: str
    pi_upper_bound_unit: str
    diss_pi_gain_unit: str
    diss_pi_int_cutoff_frequency_unit: str
    diss_pi_lower_bound_unit: str
    diss_pi_upper_bound_unit: str

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> PllInfo:
        amplitude_control = cursor.read_u32_le()
        drive_amplitude = cursor.read_f64_le()
        drive_ref_frequency = cursor.read_f64_le()
        lockin_freq_offset = cursor.read_f64_le()
        lockin_harmonic_factor = cursor.read_f64_le()
        lockin_phase_offset = cursor.read_f64_le()
        pi_gain = cursor.read_f64_le()
        pi_int_cutoff_frequency = cursor.read_f64_le()
        pi_lower_bound = cursor.read_f64_le()
        pi_upper_bound = cursor.read_f64_le()
        diss_pi_gain = cursor.read_f64_le()
        diss_pi_int_cutoff_frequency = cursor.read_f64_le()
        diss_pi_lower_bound = cursor.read_f64_le()
        diss_pi_upper_bound = cursor.read_f64_le()

        lockin_filter_cutoff_frequency = cursor.read_sm4_string()

        drive_amplitude_unit = cursor.read_sm4_string()
        drive_ref_frequency_unit = cursor.read_sm4_string()
        lockin_freq_offset_unit = cursor.read_sm4_string()
        lockin_harmonic_factor_unit = cursor.read_sm4_string()
        lockin_phase_offset_unit = cursor.read_sm4_string()
        pi_gain_unit = cursor.read_sm4_string()
        pi_int_cutoff_frequency_unit = cursor.read_sm4_string()
        pi_lower_bound_unit = cursor.read_sm4_string()
        pi_upper_bound_unit = cursor.read_sm4_string()
        diss_pi_gain_unit = cursor.read_sm4_string()
        diss_pi_int_cutoff_frequency_unit = cursor.read_sm4_string()
        diss_pi_lower_bound_unit = cursor.read_sm4_string()
        diss_pi_upper_bound_unit = cursor.read_sm4_string()

        return cls(
            amplitude_control,
            drive_amplitude,
            drive_ref_frequency,
            lockin_freq_offset,
            lockin_harmonic_factor,
            lockin_phase_offset,
            pi_gain,
            pi_int_cutoff_frequency,
            pi_lower_bound,
            pi_upper_bound,
            diss_pi_gain,
            diss_pi_int_cutoff_frequency,
            diss_pi_lower_bound,
            diss_pi_upper_bound,
            lockin_filter_cutoff_frequency,
            drive_amplitude_unit,
            drive_ref_frequency_unit,
            lockin_freq_offset_unit,
            lockin_harmonic_factor_unit,
            lockin_phase_offset_unit,
            pi_gain_unit,
            pi_int_cutoff_frequency_unit,
            pi_lower_bound_unit,
            pi_upper_bound_unit,
            diss_pi_gain_unit,
            diss_pi_int_cutoff_frequency_unit,
            diss_pi_lower_bound_unit,
            diss_pi_upper_bound_unit,
        )


@dataclass
class ChannelDriveInfo:
    master_osciallator: int
    amplitude: float
    frequency: float
    phase_offset: float
    harmonic_factor: float
    amplitude_unit: str
    frequency_unit: str
    phase_offset_unit: str
    harmonic_factor_unit: str

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> ChannelDriveInfo:
        _ = cursor.read_u32_le()
        master_osciallator = cursor.read_u32_le()
        amplitude = cursor.read_f64_le()
        frequency = cursor.read_f64_le()
        phase_offset = cursor.read_f64_le()
        harmonic_factor = cursor.read_f64_le()
        amplitude_unit = cursor.read_sm4_string()
        frequency_unit = cursor.read_sm4_string()
        phase_offset_unit = cursor.read_sm4_string()
        harmonic_factor_unit = cursor.read_sm4_string()

        return cls(
            master_osciallator,
            amplitude,
            frequency,
            phase_offset,
            harmonic_factor,
            amplitude_unit,
            frequency_unit,
            phase_offset_unit,
            harmonic_factor_unit,
        )


@dataclass
class LockinInfo:
    num_strings: int
    non_master_oscillator: int
    frequency: float
    harmonic_factor: float
    phase_offset: float
    # these might be not included
    filter_cutoff_frequency: str
    frequency_unit: str
    phase_unit: str

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> LockinInfo:
        num_strings = cursor.read_u32_le()

        non_master_oscillator = cursor.read_u32_le()
        frequency = cursor.read_f64_le()
        harmonic_factor = cursor.read_f64_le()
        phase_offset = cursor.read_f64_le()
        # these might be not included
        filter_cutoff_frequency = cursor.read_sm4_string()
        frequency_unit = cursor.read_sm4_string()
        phase_unit = cursor.read_sm4_string()

        return cls(
            num_strings,
            non_master_oscillator,
            frequency,
            harmonic_factor,
            phase_offset,
            filter_cutoff_frequency,
            frequency_unit,
            phase_unit,
        )


@dataclass
class PiControllerInfo:
    setpoint: float
    proportional_gain: float
    integral_gain: float
    lower_bound: float
    upper_bound: float
    feedback_unit: str
    setpoint_unit: str
    proportional_gain_unit: str
    integral_gain_unit: str
    output_unit: str

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> PiControllerInfo:
        setpoint = cursor.read_f64_le()
        proportional_gain = cursor.read_f64_le()
        integral_gain = cursor.read_f64_le()
        lower_bound = cursor.read_f64_le()
        upper_bound = cursor.read_f64_le()
        _ = cursor.read_u32_le()
        feedback_unit = cursor.read_sm4_string()
        setpoint_unit = cursor.read_sm4_string()
        proportional_gain_unit = cursor.read_sm4_string()
        integral_gain_unit = cursor.read_sm4_string()
        output_unit = cursor.read_sm4_string()

        return cls(
            setpoint,
            proportional_gain,
            integral_gain,
            lower_bound,
            upper_bound,
            feedback_unit,
            setpoint_unit,
            proportional_gain_unit,
            integral_gain_unit,
            output_unit,
        )


@dataclass
class LowpassFilterInfo:
    info: str

    @classmethod
    def from_buffer(cls, cursor: Cursor) -> LowpassFilterInfo:
        _ = cursor.read_u32_le()
        lowpass_filter_info = cursor.read_sm4_string()

        return cls(lowpass_filter_info)



from io import BufferedReader
import struct
from typing import cast


class Cursor:
    def __init__(self, buffer: BufferedReader):
        self._buffer = buffer

    def set_position(self, position: int) -> None:
        self._buffer.seek(position)

    def skip(self, bytes_to_skip: int) -> None:
        self.set_position(self._buffer.tell() + bytes_to_skip)

    def read(self, num_bytes: int) -> bytes:
        return self._buffer.read(num_bytes)

    def read_string(self, str_len: int) -> str:
        """Read bytes as string from _bytes with str_len"""
        return "".join(
            [chr(i).rstrip("\x00") for i in self._buffer.read(str_len)]
        )

    def read_sm4_string(self) -> str:
        length = self.read_u16_le()
        return self.read_string(length * 2)

    def read_u8_le(self) -> int:
        return cast(int, struct.unpack("<B", self._buffer.read(1))[0])

    def read_u16_le(self) -> int:
        return cast(int, struct.unpack("<H", self._buffer.read(2))[0])

    def read_i16_le(self) -> int:
        return cast(int, struct.unpack("<h", self._buffer.read(2))[0])

    def read_u32_le(self) -> int:
        return cast(int, struct.unpack("<I", self._buffer.read(4))[0])

    def read_i32_le(self) -> int:
        return cast(int, struct.unpack("<i", self._buffer.read(4))[0])

    def read_u64_le(self) -> int:
        return cast(int, struct.unpack("<q", self._buffer.read(8))[0])

    def read_f32_le(self) -> float:
        return cast(float, struct.unpack("<f", self._buffer.read(4))[0])

    def read_f64_le(self) -> float:
        return cast(float, struct.unpack("<d", self._buffer.read(8))[0])
